'use strict';

Object.defineProperty(exports, '__esModule', {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _warning = require('warning');

var _warning2 = _interopRequireDefault(_warning);

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _AsyncUtils = require('./AsyncUtils');

var _RouteUtils = require('./RouteUtils');

var _ActiveUtils = require('./ActiveUtils');

var _RoutingUtils = require('./RoutingUtils');

var _PropTypes = require('./PropTypes');

var _Location = require('./Location');

var _Location2 = _interopRequireDefault(_Location);

var _React$PropTypes = _react2['default'].PropTypes;
var any = _React$PropTypes.any;
var array = _React$PropTypes.array;
var func = _React$PropTypes.func;
var object = _React$PropTypes.object;
var instanceOf = _React$PropTypes.instanceOf;

var ContextMixin = {

  childContextTypes: {
    router: object.isRequired
  },

  getChildContext: function getChildContext() {
    return {
      router: this
    };
  },

  makePath: function makePath(pathname, query) {
    return this.props.history.makePath(pathname, query);
  },

  makeHref: function makeHref(pathname, query) {
    return this.props.history.makeHref(pathname, query);
  },

  transitionTo: function transitionTo(pathname, query) {
    var state = arguments[2] === undefined ? null : arguments[2];
    var history = this.props.history;

    var path = this.makePath(pathname, query);

    if (this.nextLocation) {
      history.replaceState(state, path);
    } else {
      history.pushState(state, path);
    }
  },

  replaceWith: function replaceWith(pathname, query) {
    var state = arguments[2] === undefined ? null : arguments[2];
    var history = this.props.history;

    var path = this.makePath(pathname, query);

    history.replaceState(state, path);
  },

  go: function go(n) {
    this.props.history.go(n);
  },

  goBack: function goBack() {
    this.go(-1);
  },

  goForward: function goForward() {
    this.go(1);
  },

  isActive: function isActive(pathname, query) {
    var location = this.state.location;

    if (location == null) return false;

    return (0, _ActiveUtils.pathnameIsActive)(pathname, location.pathname) && (0, _ActiveUtils.queryIsActive)(query, location.query);
  }

};

var Router = _react2['default'].createClass({
  displayName: 'Router',

  mixins: [ContextMixin],

  statics: {

    match: function match(routes, history, callback) {}

  },

  propTypes: {
    history: _PropTypes.history.isRequired,
    children: _PropTypes.routes,
    routes: _PropTypes.routes, // Alias for children
    createElement: func.isRequired,
    onError: func.isRequired,
    onUpdate: func,

    // For server-side rendering
    location: any,
    branch: _PropTypes.routes,
    params: object,
    components: _PropTypes.components
  },

  getDefaultProps: function getDefaultProps() {
    return {
      createElement: _react.createElement,
      onError: function onError(error) {
        // Throw errors by default so we don't silently swallow them!
        throw error; // This error probably originated in getChildRoutes or getComponents.
      }
    };
  },

  getInitialState: function getInitialState() {
    return {
      location: null,
      branch: null,
      params: null,
      components: null,
      isTransitioning: false
    };
  },

  _updateState: function _updateState(location) {
    var _this = this;

    (0, _invariant2['default'])(_Location2['default'].isLocation(location), 'A <Router> needs a valid Location');

    this.nextLocation = location;
    this.setState({ isTransitioning: true });

    this._getState(this.routes, location, function (error, state) {
      if (error || _this.nextLocation !== location) {
        _this._finishTransition(error);
      } else if (state == null) {
        (0, _warning2['default'])(false, 'Location "%s" did not match any routes', location.pathname);
        _this._finishTransition();
      } else {
        state.location = location;

        _this._runTransitionHooks(state, function (error) {
          if (error || _this.nextLocation !== location) {
            _this._finishTransition(error);
          } else {
            _this._getComponents(state, function (error, components) {
              if (error || _this.nextLocation !== location) {
                _this._finishTransition(error);
              } else {
                state.components = components;

                _this._finishTransition(null, state);
              }
            });
          }
        });
      }
    });
  },

  _finishTransition: function _finishTransition(error, state) {
    this.setState({ isTransitioning: false });
    this.nextLocation = null;

    if (error) {
      this.handleError(error);
    } else if (state) {
      this.setState(state, this.props.onUpdate);
      this._alreadyUpdated = true;
    }
  },

  _getState: function _getState(routes, location, callback) {
    var _props = this.props;
    var branch = _props.branch;
    var params = _props.params;

    if (branch && params && query) {
      callback(null, { branch: branch, params: params });
    } else {
      (0, _RoutingUtils.getState)(routes, location, callback);
    }
  },

  _runTransitionHooks: function _runTransitionHooks(nextState, callback) {
    var _this2 = this;

    // Run component hooks before route hooks.
    var hooks = this.transitionHooks.map(function (hook) {
      return (0, _RoutingUtils.createTransitionHook)(hook, _this2);
    });

    hooks.push.apply(hooks, (0, _RoutingUtils.getTransitionHooks)(this.state, nextState));

    var nextLocation = this.nextLocation;

    (0, _AsyncUtils.loopAsync)(hooks.length, function (index, next, done) {
      var hook = hooks[index];

      hooks[index].call(_this2, nextState, _this2, function (error) {
        if (error || _this2.nextLocation !== nextLocation) {
          done.call(_this2, error); // No need to continue.
        } else {
          next.call(_this2);
        }
      });
    }, callback);
  },

  _getComponents: function _getComponents(nextState, callback) {
    if (this.props.components) {
      callback(null, this.props.components);
    } else {
      (0, _RoutingUtils.getComponents)(nextState, callback);
    }
  },

  _createElement: function _createElement(component, props) {
    return typeof component === 'function' ? this.props.createElement(component, props) : null;
  },

  /**
   * Adds a transition hook that runs before all route hooks in a
   * transition. The signature is the same as route transition hooks.
   */
  addTransitionHook: function addTransitionHook(hook) {
    this.transitionHooks.push(hook);
  },

  /**
   * Removes the given transition hook.
   */
  removeTransitionHook: function removeTransitionHook(hook) {
    this.transitionHooks = this.transitionHooks.filter(function (h) {
      return h !== hook;
    });
  },

  /**
   * Cancels the current transition, preventing any subsequent transition
   * hooks from running and restoring the previous location.
   */
  cancelTransition: function cancelTransition() {
    (0, _invariant2['default'])(this.state.location, 'Router#cancelTransition: You may not cancel the initial transition');

    if (this.nextLocation) {
      this.nextLocation = null;

      // The best we can do here is goBack so the location state reverts
      // to what it was. However, we also set a flag so that we know not
      // to run through _updateState again.
      this._ignoreNextHistoryChange = true;
      this.goBack();
    } else {
      (0, _warning2['default'])(false, 'Router#cancelTransition: Router is not transitioning');
    }
  },

  handleHistoryChange: function handleHistoryChange() {
    if (this._ignoreNextHistoryChange) {
      this._ignoreNextHistoryChange = false;
    } else {
      this._updateState(this.props.history.location);
    }
  },

  componentWillMount: function componentWillMount() {
    var _props2 = this.props;
    var history = _props2.history;
    var routes = _props2.routes;
    var children = _props2.children;

    (0, _invariant2['default'])(routes || children, 'A <Router> needs some routes');

    this.routes = (0, _RouteUtils.createRoutes)(routes || children);
    this.transitionHooks = [];
    this.nextLocation = null;

    if (typeof history.setup === 'function') history.setup();

    // We need to listen first in case we redirect immediately.
    if (history.addChangeListener) history.addChangeListener(this.handleHistoryChange);

    this._updateState(history.location);
  },

  componentDidMount: function componentDidMount() {
    // React doesn't fire the setState callback when we call setState
    // synchronously within componentWillMount, so we need this. Note
    // that we still only get one call to onUpdate, even if setState
    // was called multiple times in componentWillMount.
    if (this._alreadyUpdated && this.props.onUpdate) this.props.onUpdate.call(this);
  },

  componentWillReceiveProps: function componentWillReceiveProps(nextProps) {
    (0, _invariant2['default'])(this.props.history === nextProps.history, '<Router history> may not be changed');

    var currentRoutes = this.props.routes || this.props.children;
    var nextRoutes = nextProps.routes || nextProps.children;

    if (currentRoutes !== nextRoutes) {
      this.routes = (0, _RouteUtils.createRoutes)(nextRoutes);

      // Call this here because _updateState
      // uses this.routes to determine state.
      if (nextProps.history.location) this._updateState(nextProps.history.location);
    }
  },

  componentWillUnmount: function componentWillUnmount() {
    var history = this.props.history;

    if (history.removeChangeListener) history.removeChangeListener(this.handleHistoryChange);
  },

  render: function render() {
    var _this3 = this;

    var _state = this.state;
    var location = _state.location;
    var branch = _state.branch;
    var params = _state.params;
    var components = _state.components;
    var isTransitioning = _state.isTransitioning;

    var element = null;

    if (components) {
      element = components.reduceRight(function (element, components, index) {
        if (components == null) return element; // Don't create new children; use the grandchildren.

        var route = branch[index];
        var routeParams = (0, _RoutingUtils.getRouteParams)(route, params);
        var props = { location: location, params: params, route: route, routeParams: routeParams, isTransitioning: isTransitioning };

        if ((0, _react.isValidElement)(element)) {
          props.children = element;
        } else if (element) {
          // In render, do var { header, sidebar } = this.props;
          _extends(props, element);
        }

        if (typeof components === 'object') {
          var elements = {};

          for (var key in components) if (components.hasOwnProperty(key)) elements[key] = _this3._createElement(components[key], props);

          return elements;
        }

        return _this3._createElement(components, props);
      }, element);
    }

    (0, _invariant2['default'])(element === null || element === false || (0, _react.isValidElement)(element), 'The root route must render a single element');

    return element;
  }

});

exports.Router = Router;
exports['default'] = Router;

// TODO: Mimic what we're doing in _updateState, but statically
// so we can get the right props for doing server-side rendering.